import threading
from time import time

from karabo.bound import KARABO_CLASSINFO
from PipeToZeroMQ import PipeToZeroMQ, conversion, device_schema

from . import shmem_utils

from ._version import version as deviceVersion


@KARABO_CLASSINFO("ShmemToZMQ", deviceVersion)
class ShmemToZMQ(PipeToZeroMQ.PipeToZeroMQ):
    def initialization(self):
        super().initialization()
        self._source_to_shmem_ary = {}
        self._source_to_shmem_mem = {}
        self._buffer_lock = threading.Lock()

    def _get_shmem_buffer_data(self, source, shmem_handle):
        # TODO: handle failure if this was not a shmem handle
        name, dtype, shape, index = shmem_utils.parse_shmem_handle(shmem_handle)
        with self._buffer_lock:
            # may have to open shared memory buffer
            if source not in self._source_to_shmem_ary:
                self.log.INFO(f"Opening buffer {name} for source {source}")
                try:
                    mem, ary = shmem_utils.open_shmem_from_handle(shmem_handle)
                except OSError:
                    self.log.WARN(f"Failed to open buffer {name}")
                    return None
                self._source_to_shmem_mem[source] = mem
                self._source_to_shmem_ary[source] = ary
            elif self._source_to_shmem_ary[source].shape != shape:
                self.log.INFO(f"Updating buffer shape for {source} to {shape}")
                self._source_to_shmem_ary[source] = self._source_to_shmem_mem[
                    source
                ].ndarray(shape=shape, dtype=dtype)

            # grab data from shared memory buffer
            return self._source_to_shmem_ary[source][index]

    def onInput(self, input_channel):
        actual = self.getActualTimestamp()
        input_tic = time()
        self.info["inputUpdated"] += 1
        self.info["dataRecv"] = input_channel.size()
        all_meta = input_channel.getMetaData()

        data = {}
        meta = {}

        for idx in range(input_channel.size()):
            # Read metadata
            metadata = self._extract_metadata(all_meta, idx)
            source = metadata["source"]
            if source not in self.sources:
                self.appendSchema(device_schema.timestamp_schema(source))
                self.sources.add(source)
            self._time_info(metadata, actual, self.info)

            # Read data
            hash_data = input_channel.read(idx)

            # filters
            if self.allowed_sources and source not in self.allowed_sources:
                continue
            forward, ignore = self._filter_properties(hash_data, source)

            meta[source] = conversion.meta_to_dict(metadata)
            meta[source]["ignored_keys"] = ignore
            # only this bit differs from PipeToZeroMQ.onInput
            dic, arr = conversion.hash_to_dict(
                hash_data, paths=forward, version=self.version
            )
            for shmem_handle_path in dic.pop("calngShmemPaths", []):
                shmem_handle = dic.pop(shmem_handle_path, None)
                if shmem_handle_path is None:
                    self.log.INFO(
                        f"Hash from {source} did not have {shmem_handle_path}"
                    )
                    continue
                actual_data = self._get_shmem_buffer_data(source, shmem_handle)
                arr[shmem_handle_path] = actual_data

            data[source] = (dic, arr)

        # forward data to all connected ZMQ sockets
        self._send(data, meta)

        output_tic = time()
        self.info["onInputTotal"] = 1000 * (output_tic - input_tic)

        # update properties on device
        self._updateProperties(output_tic)
        # block if device is in passive state
        self.monitoring.wait()

    def preDestruction(self):
        for ary in self._source_to_shmem_ary.values():
            del ary
        super().preDestruction()
